"""
Probability Distribution.

This module provides a class implementing a probability distribution along with its properties.
"""

import numpy as np
import scipy.stats
import scipy.sparse
import scipy._lib._util


class Distribution:
    """
    A class representing probability distributions.

    This class is primarily intended to be subclassed to provide
    distribution-specific implementations of the various methods
    (``logpdf``, ``logcdf``, ``sample``, ``mean``, ``var``, ...). When
    creating a subclass implementing a certain distribution for which
    operations like addition or multiplication are available, please
    consider implementing them, as done for instance in :class:`Normal`.
    The overriden methods should also update the ``parameters`` of the
    distribution. This allows arithmetic operations on instances of
    :class:`Distribution` and :class:`RandomVariable`.

    Parameters
    ----------
    parameters : dict
        Dictionary of distribution parameters such as mean, variance, shape, etc.
    pdf : callable
        Probability density or mass function.
    logpdf : callable
        Log-probability density or mass function.
    cdf : callable
        Cumulative distribution function.
    logcdf : callable
        Log-cumulative distribution function.
    sample : callable
        Function implementing sampling. Must have signature ``sample(size=())``.
    mean : callable
        Function returning the mean of the distribution.
    cov : callable
        Function returning the covariance of the distribution.
    shape : tuple
        Shape of samples from this distribution.
    dtype : numpy.dtype or object
        Data type of realizations of a random variable with this distribution. If ``object`` will be converted to ``numpy.dtype``.
    random_state : None or int or :class:`~numpy.random.RandomState` instance, optional
        This parameter defines the RandomState object to use for drawing
        realizations from this distribution.
        If None (or np.random), the global np.random state is used.
        If integer, it is used to seed the local :class:`~numpy.random.RandomState` instance.
        Default is None.

    See Also
    --------
    RandomVariable : Random variables are the main objects used by probabilistic numerical methods.

    Examples
    --------

    """

    def __init__(
        self,
        parameters=None,
        pdf=None,
        logpdf=None,
        cdf=None,
        logcdf=None,
        sample=None,
        mean=None,
        cov=None,
        shape=None,
        dtype=None,
        random_state=None,
    ):
        if parameters is None:
            parameters = {}  # sentinel value to avoid anti-pattern
        self._parameters = parameters
        self._pdf = pdf
        self._logpdf = logpdf
        self._cdf = cdf
        self._logcdf = logcdf
        self._sample = sample
        self._mean = mean
        self._cov = cov
        self._set_shape(shape)
        self._dtype = dtype
        self._random_state = scipy._lib._util.check_random_state(random_state)

    def _check_distparams(self):
        pass
        # TODO: type checking of self._parameters; check method signatures.

    def _set_shape(self, shape):
        """
        Sets shape in accordance with distribution mean.
        """
        self._shape = shape
        try:
            # Set shape based on mean
            if np.isscalar(self.mean()):
                shape_mean = ()
            else:
                shape_mean = self.mean().shape
            if shape is None or shape_mean == shape:
                self._shape = shape_mean
            else:
                raise ValueError(
                    "Shape of distribution mean and given shape do not match."
                )
        except NotImplementedError:
            # Set shape based on a sample
            if np.isscalar(self.sample(size=1)):
                shape_sample = ()
            else:
                shape_sample = self.sample(size=1).shape
            if shape is None or shape_sample == shape:
                self._shape = shape_sample
            else:
                raise ValueError(
                    "Shape of distribution sample and given shape do not match."
                )

    @property
    def shape(self):
        """Shape of samples from this distribution."""
        return self._shape

    @property
    def dtype(self):
        """``Dtype`` of elements of samples from this distribution."""
        return self._dtype

    @dtype.setter
    def dtype(self, newtype):
        """Set the ``dtype`` of the distribution."""
        self._dtype = newtype

    @property
    def random_state(self):
        """Random state of the distribution.

        This attribute defines the RandomState object to use for drawing
        realizations from this distribution.
        If None (or np.random), the global np.random state is used.
        If integer, it is used to seed the local :class:`~numpy.random.RandomState` instance.
        """
        return self._random_state

    @random_state.setter
    def random_state(self, seed):
        """ Get or set the RandomState object of the underlying distribution.

        This can be either None or an existing RandomState object.
        If None (or np.random), use the RandomState singleton used by np.random.
        If already a RandomState instance, use it.
        If an int, use a new RandomState instance seeded with seed.
        """
        self._random_state = scipy._lib._util.check_random_state(seed)

    @property
    def parameters(self):
        """
        Parameters of the probability distribution.

        The parameters of the distribution such as mean, variance, et cetera stored in a ``dict``.
        """
        if self._parameters is not None:
            return self._parameters
        else:
            raise AttributeError(
                "No parameters of {} are available.".format(type(self).__name__)
            )

    def pdf(self, x):
        """
        Probability density or mass function.

        Parameters
        ----------
        x : array-like
            Evaluation points of the probability density / mass function.

        Returns
        -------
        p : array-like
            Value of the probability density / mass function at the given points.

        """
        if self._pdf is not None:
            return self._pdf(x)
        if self._logpdf is not None:
            return np.exp(self._logpdf(x))
        raise NotImplementedError(
            "The function 'pdf' is not implemented for object of class {}".format(
                type(self).__name__
            )
        )

    def logpdf(self, x):
        """
        Natural logarithm of the probability density function.

        Parameters
        ----------
        x : array-like
            Evaluation points of the log-probability density/mass function.

        Returns
        -------
        logp : array-like
            Value of the log-probability density / mass function at the given points.
        """
        if self._logpdf is not None:
            return self._logpdf(x)
        elif self._pdf is not None:
            return np.log(self._pdf(x))
        else:
            raise NotImplementedError(
                "The function 'logpdf' is not implemented for object of class {}".format(
                    type(self).__name__
                )
            )

    def cdf(self, x):
        """
        Cumulative distribution function.

        Parameters
        ----------
        x : array-like
            Evaluation points of the cumulative distribution function.

        Returns
        -------
        q : array-like
            Value of the cumulative density function at the given points.
        """
        if self._cdf is not None:
            return self._cdf(x)
        elif self._logcdf is not None:
            return np.exp(self._logcdf(x))
        else:
            raise NotImplementedError(
                "The function 'cdf' is not implemented for object of class {}".format(
                    type(self).__name__
                )
            )

    def logcdf(self, x):
        """
        Log-cumulative distribution function.

        Parameters
        ----------
        x : array-like
            Evaluation points of the cumulative distribution function.

        Returns
        -------
        q : array-like
            Value of the log-cumulative density function at the given points.
        """
        if self._logcdf is not None:
            return self._logcdf(x)
        elif self._cdf is not None:
            return np.log(self._cdf(x))
        else:
            raise NotImplementedError(
                "The function 'logcdf' is not implemented for object of class {}".format(
                    type(self).__name__
                )
            )

    def sample(self, size=()):
        """
        Returns realizations from the associated random variable.

        Parameters
        ----------
        size : tuple, default=()
            Size of the realizations.

        Returns
        -------
        realizations : array-like or LinearOperator
            Realizations of the underlying random variable.
        """
        if self._sample is not None:
            return self._sample(size=size)
        else:
            raise NotImplementedError(
                "The function 'sample' is not implemented for object of class {}.".format(
                    type(self).__name__
                )
            )

    def median(self):
        """
        Median of the distribution.

        Returns
        -------
        median : float
            The median of the distribution.
        """
        return self.cdf(x=0.5)

    def mode(self):
        """
        Mode of the distribution.

        Returns
        -------
        mode : float
            The mode of the distribution.
        """
        if "mode" in self._parameters:
            return self._parameters["mode"]
        else:
            raise NotImplementedError(
                "The function 'mode' is not implemented for object of class {}.".format(
                    type(self).__name__
                )
            )

    def mean(self):
        """
        Mean :math:`\\mathbb{E}(X)` of the distribution.

        Returns
        -------
        mean : array-like
            The mean of the distribution.
        """
        if self._mean is not None:
            return self._mean()
        elif "mean" in self._parameters:
            return self._parameters["mean"]
        else:
            raise NotImplementedError(
                "The function 'mean' is not implemented for object of class {}.".format(
                    type(self).__name__
                )
            )

    def cov(self):
        """
        Covariance :math:`\\operatorname{Cov}(X) = \\mathbb{E}((X-\\mathbb{E}(X))(X-\\mathbb{E}(X))^\\top)` of the
        distribution.

        Returns
        -------
        cov : array-like
            The kernels of the distribution.
        """
        if self._cov is not None:
            return self._cov()
        elif "cov" in self._parameters:
            return self._parameters["cov"]
        else:
            raise NotImplementedError(
                "The function 'cov' is not implemented for object of class {}".format(
                    type(self).__name__
                )
            )

    def var(self):
        """
        Variance :math:`\\operatorname{Var}(X) = \\mathbb{E}((X-\\mathbb{E}(X))^2)` of the distribution.

        Returns
        -------
        var : array-like
            The variance of the distribution.
        """
        if "var" in self._parameters:
            return self._parameters["var"]
        else:
            raise NotImplementedError(
                "The function 'var' is not implemented for object of class {}".format(
                    type(self).__name__
                )
            )

    def std(self):
        """
        Standard deviation of the distribution.

        Returns
        -------
        std : array-like
            The standard deviation of the distribution.
        """
        return np.sqrt(self.var())

    def reshape(self, newshape):
        """
        Give a new shape to (realizations of) this distribution.

        Parameters
        ----------
        newshape : int or tuple of ints
            New shape for the realizations and parameters of this distribution. It must be compatible with the original
            shape.

        Returns
        -------
        reshaped_rv : ``self`` with the new dimensions of ``shape``.
        """
        raise NotImplementedError(
            "Reshaping not implemented for distribution of type: {}.".format(
                self.__class__.__name__
            )
        )



"""
Quadrature, i.e. numerical integration.

This module collects both classic and Bayesian quadrature rules used for numerical integration of functions.

Bayesian quadrature methods integrate a function by iteratively building a probabilistic model and using its predictions
to adaptively choose points to evaluate the integrand.
"""

from probnum.quad.quadrature import *
from probnum.quad.bayesian import *
from probnum.quad.polynomial import *

# Public classes and functions. Order is reflected in documentation.
__all__ = [
    "quad",
    "nquad",
    "bayesquad",
    "nbayesquad",
    "Quadrature",
    "PolynomialQuadrature",
    "BayesianQuadrature",
    "VanillaBayesianQuadrature",
    "WSABIBayesianQuadrature",
    "ClenshawCurtis",
]

# Set correct module paths. Corrects links and module paths in documentation.
Quadrature.__module__ = "probnum.quad"
BayesianQuadrature.__module__ = "probnum.quad"
PolynomialQuadrature.__module__ = "probnum.quad"



"""
Quadrature, i.e. numerical integration.

This module provides an abstract base class defining quadrature methods.
"""

import abc


def quad(fun, bounds, type=None):
    """
    One-dimensional numerical integration.

    Parameters
    ----------
    fun : function
        Function to be integrated.
    bounds : ndarray, shape=(2,)
        Domain of integration.
    type : str
        Type of quadrature to use. The available options are

        ====================  ===========
         Bayesian              ``bayes``
         Clenshaw-Curtis       ``cc``
        ====================  ===========

    Returns
    -------
    F : RandomVariable
        The integral of ``fun`` within the given bounds.
    """
    raise NotImplementedError


def nquad(fun, domain, type=None):
    """
    N-dimensional numerical integration.

    Parameters
    ----------
    fun : function
        Function to be integrated.
    domain : ndarray, shape=(d,2)
        Domain of integration.
    type : str
        Type of quadrature to use. The available options are

        ====================  ===========
         Bayesian              ``bayes``
         Clenshaw-Curtis       ``cc``
        ====================  ===========

    Returns
    -------
    F : RandomVariable
        The integral of ``fun`` on the domain.
    """
    raise NotImplementedError


class Quadrature(abc.ABC):
    """
    An abstract base class for quadrature methods.

    This class is designed to be subclassed by quadrature implementations.
    """

    def __init__(self):
        """
        """

    def integrate(self, fun, **kwargs):
        """
        Numerically integrate the given function.

        Parameters
        ----------
        fun : function
            Function to be integrated.
        kwargs

        Returns
        -------

        """
        raise NotImplementedError



from .arrayutils import *
from .fctutils import *

# Public classes and functions. Order is reflected in documentation.
__all__ = [
    "atleast_1d",
    "atleast_2d",
    "as_colvec",
    "assert_is_1d_ndarray",
    "assert_is_2d_ndarray",
    "assert_evaluates_to_scalar",
]



"""
Dirac delta distribution.
"""
import operator

import numpy as np

from probnum.prob.distributions.distribution import Distribution


class Dirac(Distribution):
    """
    The Dirac delta distribution.

    This distribution models a point mass and can be useful to represent
    numbers as random variables with Dirac measure. It has the useful
    property that arithmetic operations between a :class:`Dirac` random
    variable and an arbitrary :class:`RandomVariable` acts in the same
    way as the arithmetic operation with a constant.

    Note, that a Dirac measure does not admit a probability density
    function but can be viewed as a distribution (generalized function).

    Parameters
    ----------
    support : scalar or array-like or LinearOperator
        The support of the dirac delta function.

    See Also
    --------
    Distribution : Class representing general probability distribution.

    Examples
    --------
    >>> from probnum.prob import RandomVariable, Dirac
    >>> dist1 = Dirac(support=0.)
    >>> dist2 = Dirac(support=1.)
    >>> rv = RandomVariable(distribution=dist1 + dist2)
    >>> rv.sample(size=5)
    array([1., 1., 1., 1., 1.])
    """

    def __init__(self, support, random_state=None):
        if np.isscalar(support):
            _dtype = np.dtype(type(support))
        else:
            _dtype = support.dtype
        super().__init__(
            parameters={"support": support}, dtype=_dtype, random_state=random_state
        )

    def cdf(self, x):
        if np.any(x < self.parameters["support"]):
            return 0.0
        else:
            return 1.0

    def median(self):
        return self.parameters["support"]

    def mode(self):
        return self.parameters["support"]

    def mean(self):
        return self.parameters["support"]

    def var(self):
        return 0.0

    def cov(self):
        if np.isscalar(self.parameters["support"]):
            return self.var()
        else:
            return np.zeros(
                (len(self.parameters["support"]), len(self.parameters["support"]))
            )

    def sample(self, size=(), seed=None):
        ndims = len(self.shape)
        if size == 1 or size == ():
            return self.parameters["support"]
        elif isinstance(size, int) and ndims == 0:
            return np.tile(A=self.parameters["support"], reps=size)
        elif isinstance(size, int):
            return np.tile(
                A=self.parameters["support"], reps=[size, *np.repeat(1, ndims)]
            )
        else:
            return np.tile(
                A=self.parameters["support"], reps=tuple([*size, *np.repeat(1, ndims)])
            )

    def reshape(self, newshape):
        try:
            # Reshape support
            self._parameters["support"].reshape(newshape=newshape)
        except ValueError:
            raise ValueError(
                "Cannot reshape this Dirac distribution to the given shape: {}".format(
                    str(newshape)
                )
            )

    # Binary arithmetic operations
    def __add__(self, other):
        if isinstance(other, Dirac):
            support_ = self.parameters["support"] + other.parameters["support"]
            return Dirac(support=support_, random_state=self.random_state)
        else:
            return other.__add__(other=self)

    def __sub__(self, other):
        if isinstance(other, Dirac):
            support_ = self.parameters["support"] - other.parameters["support"]
            return Dirac(support=support_, random_state=self.random_state)
        else:
            return other.__rsub__(other=self)

    def __mul__(self, other):
        if isinstance(other, Dirac):
            support_ = self.parameters["support"] * other.parameters["support"]
            return Dirac(support=support_, random_state=self.random_state)
        else:
            return other.__mul__(other=self)

    def __matmul__(self, other):
        if isinstance(other, Dirac):
            support_ = self.parameters["support"] @ other.parameters["support"]
            return Dirac(support=support_, random_state=self.random_state)
        else:
            return other.__rmatmul__(other=self)

    def __truediv__(self, other):
        if isinstance(other, Dirac):
            support_ = operator.truediv(
                self.parameters["support"], other.parameters["support"]
            )
            return Dirac(support=support_, random_state=self.random_state)
        else:
            return other.__rtruediv__(other=self)

    def __pow__(self, power, modulo=None):
        if isinstance(power, Dirac):
            support_ = pow(
                self.parameters["support"], power.parameters["support"], modulo
            )
            return Dirac(support=support_, random_state=self.random_state)
        else:
            return power.__rpow__(power=self, modulo=modulo)

    # Binary arithmetic operations with reflected (swapped) operands

    def __radd__(self, other):
        return other.__add__(other=self)

    def __rsub__(self, other):
        return other.__sub__(other=self)

    def __rmul__(self, other):
        return other.__mul__(other=self)

    def __rmatmul__(self, other):
        return other.__matmul__(other=self)

    def __rtruediv__(self, other):
        return other.__truediv__(other=self)

    def __rpow__(self, power, modulo=None):
        return power.__pow__(power=self)

    # Augmented arithmetic assignments (+=, -=, *=, ...)
    # attempting to do the operation in place

    def __iadd__(self, other):
        if isinstance(other, Dirac):
            support_ = self.parameters["support"] + other.parameters["support"]
            self.parameters["support"] = support_
            return self
        else:
            return NotImplemented

    def __isub__(self, other):
        if isinstance(other, Dirac):
            support_ = self.parameters["support"] - other.parameters["support"]
            self.parameters["support"] = support_
            return self
        else:
            return NotImplemented

    def __imul__(self, other):
        if isinstance(other, Dirac):
            support_ = self.parameters["support"] * other.parameters["support"]
            self.parameters["support"] = support_
            return self
        else:
            return NotImplemented

    def __imatmul__(self, other):
        if isinstance(other, Dirac):
            support_ = self.parameters["support"] @ other.parameters["support"]
            self.parameters["support"] = support_
            return self
        else:
            return NotImplemented

    def __itruediv__(self, other):
        if isinstance(other, Dirac):
            support_ = operator.truediv(
                self.parameters["support"], other.parameters["support"]
            )
            self.parameters["support"] = support_
            return self
        else:
            return NotImplemented

    def __ipow__(self, power, modulo=None):
        if isinstance(power, Dirac):
            support_ = pow(
                self.parameters["support"], power.parameters["support"], modulo
            )
            self.parameters["support"] = support_
            return self
        else:
            return NotImplemented

    # Unary arithmetic operations

    def __neg__(self):
        self.parameters["support"] = operator.neg(self.parameters["support"])
        return self

    def __pos__(self):
        self.parameters["support"] = operator.pos(self.parameters["support"])
        return self

    def __abs__(self):
        self.parameters["support"] = operator.abs(self.parameters["support"])
        return self

    def __invert__(self):
        support_ = self.parameters["support"]
        self.parameters["support"] = operator.invert(support_)
        return self



"""
Import convenience functions in optim.py to create an
intuitive, numpy-like interface.

Note
----
Local import, because with a global import this does not seem
to work.
"""
from .linearsdemodel import *
from .continuousmodel import *

__all__ = ["ContinuousModel", "LinearSDEModel", "LTISDEModel"]



"""

See BFaS; p. 84f.
"""

import numpy as np


__all__ = ["UnscentedTransform"]


class UnscentedTransform:
    """
    Used for unscented Kalman filter.
    """

    def __init__(self, ndim, spread=1e-4, priorpar=2.0, special_scale=0.0):
        """
        ndim : int
            Spatial dimensionality
        spread : float
            Spread of the sigma points around mean
        priorpar : float
            Incorporate prior knowledge about distribution of x.
            For Gaussians, 2.0 is optimal (see link below)
        special_scale : float
            Secondary scaling parameter.
            The primary parameter is computed below.

        Source
        ------
        P. 7 ("Unscented transform:") of
        https://www.pdx.edu/biomedical-signal-processing-lab/sites/www.pdx.edu.biomedical-signal-processing-lab/files/ukf.wan_.chapt7_.pdf
        """
        self.scale = _compute_scale(ndim, spread, special_scale)
        self.ndim = ndim
        self.mweights, self.cweights = self.unscented_weights(spread, priorpar)

    def unscented_weights(self, alp, bet):
        """
        See BFaS; p. 84.

        Parameters
        ----------
        alp: float
            Spread of sigma points around mean (alpha)
        bet: float
            Prior information parameter (beta)

        Returns
        -------
        np.ndarray, shape (2 * ndim + 1,)
            constant mean weights.
        np.ndarray, shape (2 * ndim + 1,)
            constant kernels weights.
        """
        mweights = _meanweights(self.ndim, self.scale)
        cweights = _covarweights(self.ndim, alp, bet, self.scale)
        return mweights, cweights

    def sigma_points(self, mean, covar):
        """
        Sigma points.

        Parameters
        ----------
        mean: np.ndarray, shape (d,)
            mean of Gaussian distribution
        covar: np.ndarray, shape (d, d)
            kernels of Gaussian distribution

        Returns
        -------
        np.ndarray, shape (2 * d + 1, d)
        """
        if len(mean) != self.ndim:
            raise ValueError("Dimensionality does not match UT")
        sigpts = np.zeros((2 * self.ndim + 1, self.ndim))
        sqrtcovar = np.linalg.cholesky(covar)
        sigpts[0] = mean.copy()
        for idx in range(self.ndim):
            sigpts[idx + 1] = mean + np.sqrt(self.ndim + self.scale) * sqrtcovar[:, idx]
            sigpts[self.ndim + 1 + idx] = (
                mean - np.sqrt(self.ndim + self.scale) * sqrtcovar[:, idx]
            )
        return sigpts

    def propagate(self, time, sigmapts, modelfct):
        """
        Propagate sigma points.

        Parameters
        ----------
        time : float
            Time :math:`t` which is passed on to the modelfunction.
        sigmapts : np.ndarray, shape=(2 N+1, N)
            Sigma points (N is the spatial dimension of the dynamic model)
        modelfct : callable, signature=(t, x, \*\*kwargs)
            Function through which to propagate

        Returns
        -------
        np.ndarray, shape=(2 N + 1, M),
            M is the dimension of the measurement model
        """
        propsigpts = np.array([modelfct(time, pt) for pt in sigmapts])
        return propsigpts

    def estimate_statistics(self, proppts, sigpts, covmat, mpred):
        """
        Computes predicted summary statistics,
        predicted mean/kernels/crosscovariance,
        from (propagated) sigmapoints.

        Not to be confused with mean and kernels resulting
        from the prediction step of the Bayesian filter.
        Hence we call it "estimate_*" instead of "predict_*".
        """
        estmean = _estimate_mean(self.mweights, proppts)
        estcovar = _estimate_covar(self.cweights, proppts, estmean, covmat)
        estcrosscovar = _estimate_crosscovar(
            self.cweights, proppts, estmean, sigpts, mpred
        )
        return estmean, estcovar, estcrosscovar


def _compute_scale(ndim, alp, kap):
    """
    See BFaS; p. 83

    Parameters
    ----------
    ndim: int
        Spatial dimensionality of state space model
    alp: float
        Spread of sigma points around mean (1; alpha)
    kap: float
        Spread of sigma points around mean (2; kappa)

    Returns
    -------
    float
        Scaling parameter for unscented transform
    """
    return alp ** 2 * (ndim + kap) - ndim


def _meanweights(ndim, lam):
    """
    Parameters
    ----------
    ndim: int
        Spatial dimensionality of state space model
    lam: float
        Scaling parameter for unscented transform (lambda)

    Returns
    -------
    np.ndarray, shape (2*ndim+1,)
        Constant mean weights.
    """
    mw0 = np.ones(1) * lam / (ndim + lam)
    mw = np.ones(2 * ndim) / (2.0 * (ndim + lam))
    return np.hstack((mw0, mw))


def _covarweights(ndim, alp, bet, lam):
    """
    Parameters
    ----------
    ndim: int
        Spatial dimensionality of state space model
    alp: float
        Spread of sigma points around mean (alpha)
    bet: float
        Prior information parameter (beta)
    lam: float
        Scaling parameter for unscented transform (lambda)

    Returns
    -------
    np.ndarray, shape (2 * ndim + 1,)
        the constant kernels weights.
    """
    cw0 = np.ones(1) * lam / (ndim + lam) + (1 - alp ** 2 + bet)
    cw = np.ones(2 * ndim) / (2.0 * (ndim + lam))
    return np.hstack((cw0, cw))


def _estimate_mean(mweights, proppts):
    """
    See BFaS; p. 88.

    Arguments
    ---------
    mweights: np.ndarray, shape (2*ndim + 1,)
        Constant mean weights for unscented transform.
    proppts: np.ndarray, shape (2*ndim + 1, ndim)
        Propagated sigma points

    Returns
    -------
    np.ndarray, shape (ndim,)
        Estimated mean.
    """
    return mweights @ proppts


def _estimate_covar(cweights, proppts, mean, covmat):
    """
    See BFaS; p. 88.

    Arguments
    ---------
    cweights: np.ndarray, shape (2*ndim + 1,)
        Constant kernels weights for unscented transform.
    proppts: np.ndarray, shape (2*ndim + 1, ndim)
        Propagated sigma points
    mean: np.ndarray, shape (ndim,)
        Result of _estimate_mean(...)
    covmat: np.ndarray, shape (ndim, ndim)
        Covariance of measurement model at current time.

    Returns
    -------
    np.ndarray, shape (ndim, ndim)
        Estimated kernels.
    """
    cent = proppts - mean
    empcov = cent.T @ (cweights * cent.T).T
    return empcov + covmat


def _estimate_crosscovar(cweights, proppts, mean, sigpts, mpred):
    """
    See BFaS; p.88.

    Arguments
    ---------
    cweights: np.ndarray, shape (2*ndim + 1,)
        Constant kernels weights for unscented transform.
    sigpts: np.ndarray, shape (2*ndim + 1, ndim)
        Sigma points
    mpred: np.ndarray, shape (ndim,)
        Predicted mean
    proppts: np.ndarray, shape (2*ndim + 1, ndim)
        Propagated sigma points
    mean: np.ndarray, shape (ndim,)
        Result of _estimate_mean(...)

    Returns
    -------
    np.ndarray, shape (ndim,)
        Estimated kernels.
    """
    cent_prop = proppts - mean
    cent_sig = sigpts - mpred
    empcrosscov = cent_sig.T @ (cweights * cent_prop.T).T
    return empcrosscov



"""
Time continuous Gauss-Markov models implicitly defined
through being a solution to the SDE
dx(t) = F(t) x(t) dt + L(t) dB(t).

If initial condition is Gaussian RV, the solution
is a Gauss-Markov process.
"""

import numpy as np
import scipy.linalg
from probnum.filtsmooth.statespace.continuous import continuousmodel
from probnum.prob import RandomVariable
from probnum.prob.distributions import Normal

__all__ = ["LinearSDEModel", "LTISDEModel"]


class LinearSDEModel(continuousmodel.ContinuousModel):
    """
    Linear time-continuous Markov models given by the solution of the
    stochastic differential equation
    :math:`dx = [F(t) x(t) + u(t)] dt + L(t) dB(t)`.


    Parameters
    ----------
    driftmatrixfct : callable, signature=(t, \*\*kwargs)
        This is F = F(t). The evaluations of this function are called
        the drift(matrix) of the SDE.
        Returns np.ndarray with shape=(n, n)
    forcfct : callable, signature=(t, \*\*kwargs)
        This is u = u(t). Evaluations of this function are called
        the force(vector) of the SDE.
        Returns np.ndarray with shape=(n,)
    dispmatrixfct : callable, signature=(t, \*\*kwargs)
        This is L = L(t). Evaluations of this function are called
        the dispersion(matrix) of the SDE.
        Returns np.ndarray with shape=(n, s)
    diffmatrix : np.ndarray, shape=(s, s)
        This is the diffusion matrix Q of the Brownian motion.
        It is always a square matrix and the size of this matrix matches
        the number of columns of the dispersionmatrix.

    Notes
    -----
    If initial conditions are Gaussian, the solution is a Gauss-Markov process.
    We assume Gaussianity for :meth:`chapmankolmogorov`.
    """

    def __init__(self, driftmatrixfct, forcfct, dispmatrixfct, diffmatrix):
        self._driftmatrixfct = driftmatrixfct
        self._forcefct = forcfct
        self._dispmatrixfct = dispmatrixfct
        self._diffmatrix = diffmatrix

    def drift(self, time, state, **kwargs):
        """
        Evaluates f(t, x(t)) = F(t) x(t) + u(t).
        """
        driftmatrix = self._driftmatrixfct(time, **kwargs)
        force = self._forcefct(time, **kwargs)
        return driftmatrix @ state + force

    def dispersion(self, time, state, **kwargs):
        """
        Evaluates l(t, x(t)) = L(t).
        """
        return self._dispmatrixfct(time, **kwargs)

    def jacobian(self, time, state, **kwargs):
        """
        maps t -> F(t)
        """
        return self._driftmatrixfct(time, **kwargs)

    @property
    def diffusionmatrix(self):
        """
        Evaluates Q.
        """
        return self._diffmatrix

    @property
    def ndim(self):
        """
        Spatial dimension (utility attribute).
        """
        return len(self._driftmatrixfct(0.0))

    def chapmankolmogorov(self, start, stop, step, randvar, **kwargs):
        """
        Solves differential equations for mean and
        kernels of the SDE solution (Eq. 5.50 and 5.51
        or Eq. 10.73 in Applied SDEs).

        By default, we assume that ``randvar`` is Gaussian.
        """
        if not issubclass(type(randvar.distribution), Normal):
            errormsg = (
                "Closed form solution for Chapman-Kolmogorov "
                "equations in linear SDE models is only "
                "available for Gaussian initial conditions."
            )
            raise ValueError(errormsg)
        mean, covar = randvar.mean(), randvar.cov()
        time = start
        while time < stop:
            meanincr, covarincr = self._increment(time, mean, covar, **kwargs)
            mean, covar = mean + step * meanincr, covar + step * covarincr
            time = time + step
        return RandomVariable(distribution=Normal(mean, covar)), None

    def _increment(self, time, mean, covar, **kwargs):
        """
        Euler step for closed form solutions of ODE defining mean
        and kernels of the solution of the Chapman-Kolmogoro
        equations (via Fokker-Planck equations, but that is not crucial
        here).
        See RHS of Eq. 10.82 in Applied SDEs.
        """
        disped = self.dispersion(time, mean, **kwargs)
        jacob = self.jacobian(time, mean, **kwargs)
        diff = self.diffusionmatrix
        newmean = self.drift(time, mean, **kwargs)
        newcovar = covar @ jacob.T + jacob @ covar.T + disped @ diff @ disped.T
        return newmean, newcovar


class LTISDEModel(LinearSDEModel):
    """
    Linear time-invariant continuous Markov models of the
    form
    dx = [F x(t) + u] dt + L dBt.
    In the language of dynamic models,
    x(t) : state process
    F : drift matrix
    u : forcing term
    L : dispersion matrix.
    Bt : Brownian motion with constant diffusion matrix Q.

    Parameters
    ----------
    driftmatrix : np.ndarray, shape=(n, n)
        This is F. It is the drift matrix of the SDE.
    force : np.ndarray, shape=(n,)
        This is U. It is the force vector of the SDE.
    dispmatrix : np.ndarray, shape(n, s)
        This is L. It is the dispersion matrix of the SDE.
    diffmatrix : np.ndarray, shape=(s, s)
        This is the diffusion matrix Q of the Brownian motion
        driving the SDE.

    Notes
    -----
    It assumes Gaussian initial conditions (otherwise
    it is no Gauss-Markov process).
    """

    def __init__(self, driftmatrix, force, dispmatrix, diffmatrix):
        """
        Parameters
        ----------
        driftmatrix : ndarray (F)
        force : ndarray (u)
        dispmatrix : ndarray (L)
        diffmatrix : ndarray (Q)
        """
        _check_initial_state_dimensions(driftmatrix, force, dispmatrix, diffmatrix)
        super().__init__(
            (lambda t, **kwargs: driftmatrix),
            (lambda t, **kwargs: force),
            (lambda t, **kwargs: dispmatrix),
            diffmatrix,
        )
        self._driftmatrix = driftmatrix
        self._force = force
        self._dispmatrix = dispmatrix
        self._diffmatrix = diffmatrix

    @property
    def driftmatrix(self):
        """
        """
        return self._driftmatrix

    @property
    def force(self):
        """
        """
        return self._force

    @property
    def dispersionmatrix(self):
        """
        """
        return self._dispmatrix

    def chapmankolmogorov(self, start, stop, step, randvar, **kwargs):
        """
        Solves Chapman-Kolmogorov equation from start to stop via step.

        For LTISDEs, there is a closed form solutions to the ODE for
        mean and kernels (see super().chapmankolmogorov(...)). We
        exploit this for [(stop - start)/step] steps.

        References
        ----------
        Eq. (8) in
        http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.390.380&rep=rep1&type=pdf
        and Eq. 6.41 and Eq. 6.42
        in Applied SDEs.
        """
        mean, cov = randvar.mean(), randvar.cov()
        if np.isscalar(mean) and np.isscalar(cov):
            mean, cov = mean * np.ones(1), cov * np.eye(1)
        increment = stop - start
        newmean = self._predict_mean(increment, mean, **kwargs)
        newcov, crosscov = self._predict_covar(increment, cov, **kwargs)
        return RandomVariable(distribution=Normal(newmean, newcov)), crosscov

    def _predict_mean(self, h, mean, **kwargs):
        """
        Predicts mean via closed-form solution to Chapman-Kolmogorov
        equation for Gauss-Markov processes according to Eq. (8) in
        http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.390.380&rep=rep1&type=pdf

        This function involves a lot of concatenation of matrices,
        hence readibility is hard to guarantee. If you know better how
        to make this readable, feedback is welcome!
        """
        drift = self.driftmatrix
        force = self.force
        extended_state = np.hstack((mean, force))
        firstrowblock = np.hstack((drift, np.eye(*drift.shape)))
        blockmat = np.hstack((firstrowblock.T, 0.0 * firstrowblock.T)).T
        proj = np.eye(*firstrowblock.shape)
        return proj @ scipy.linalg.expm(h * blockmat) @ extended_state

    def _predict_covar(self, increment, cov, **kwargs):
        """
        Predicts kernels via closed-form solution to Chapman-Kolmogorov
        equation for Gauss-Markov processes according to Eq. 6.41 and
        Eq. 6.42 in Applied SDEs.

        This function involves a lot of concatenation of matrices,
        hence readibility is hard to guarantee. If you know better how
        to make this readable, feedback is welcome!
        """
        drift = self.driftmatrix
        disp = self.dispersionmatrix
        diff = self.diffusionmatrix
        firstrowblock = np.hstack((drift, disp @ diff @ disp.T))
        secondrowblock = np.hstack((0 * drift.T, -1.0 * drift.T))
        blockmat = np.hstack((firstrowblock.T, secondrowblock.T)).T
        proj = np.eye(*firstrowblock.shape)
        initstate = np.flip(proj).T
        transformed_sol = scipy.linalg.expm(increment * blockmat) @ initstate
        trans = scipy.linalg.expm(increment * drift)
        transdiff = proj @ transformed_sol @ trans.T
        crosscov = cov @ trans.T
        newcov = trans @ crosscov + transdiff
        return newcov, crosscov


def _check_initial_state_dimensions(drift, force, disp, diff):
    """
    Checks that the matrices all align and are of proper shape.

    If all the bugs are removed and the tests run, these asserts
    are turned into Exception-catchers.

    Parameters
    ----------
    drift : np.ndarray, shape=(n, n)
    force : np.ndarray, shape=(n,)
    disp : np.ndarray, shape=(n, s)
    diff : np.ndarray, shape=(s, s)

    """
    if drift.ndim != 2 or drift.shape[0] != drift.shape[1]:
        raise ValueError("driftmatrix not of shape (n, n)")
    if force.ndim != 1:
        raise ValueError("force not of shape (n,)")
    if force.shape[0] != drift.shape[1]:
        raise ValueError("force not of shape (n,)" "or driftmatrix not of shape (n, n)")
    if disp.ndim != 2:
        raise ValueError("dispersion not of shape (n, s)")
    if diff.ndim != 2 or diff.shape[0] != diff.shape[1]:
        raise ValueError("diffusion not of shape (s, s)")
    if disp.shape[1] != diff.shape[0]:
        raise ValueError(
            "dispersion not of shape (n, s)" "or diffusion not of shape (s, s)"
        )



"""
(Finite-dimensional) Linear Operators.

This package implements a variety of finite dimensional linear operators. These have the advantage of only implementing
a matrix-vector product instead of representing the full linear operator as a matrix in memory.
"""

from probnum.linalg.linops.linearoperators import *
from probnum.linalg.linops.kronecker import *

# Public classes and functions. Order is reflected in documentation.
__all__ = [
    "LinearOperator",
    "Identity",
    "ScalarMult",
    "MatrixMult",
    "Kronecker",
    "SymmetricKronecker",
    "Vec",
    "Svec",
    "Symmetrize",
    "aslinop",
]

# Set correct module paths. Corrects links and module paths in documentation.
LinearOperator.__module__ = "probnum.linalg.linops"
ScalarMult.__module__ = "probnum.linalg.linops"



from .continuous import *
from .discrete import *
from .statespace import *

# Public classes and functions. Order is reflected in documentation.
__all__ = [
    "ContinuousModel",
    "LinearSDEModel",
    "LTISDEModel",
    "DiscreteModel",
    "DiscreteGaussianModel",
    "DiscreteGaussianLinearModel",
    "DiscreteGaussianLTIModel",
    "generate_cd",
    "generate_dd",
]



"""
"""

from abc import ABC, abstractmethod


__all__ = ["DiscreteModel"]


class DiscreteModel(ABC):
    """
    Abstract interface for state space model components.
    x(t_{i+1}) ~ p(x(t_{i+1}) | x(t_{i})).

    Nothing happens here except passing responsibilities
    of implementation down the subclasses.
    In the future, this might change so please subclass
    this object accordingly.
    """

    @abstractmethod
    def sample(self, time, state, **kwargs):
        """
        Samples x_{t} ~ p(x_{t} | x_{s})
        as a function of t and x_s (plus additional parameters).
        """
        raise NotImplementedError

    def pdf(self, loc, time, state, **kwargs):
        """
        Evaluates pdf of p(x_t | x_s).
        Required for particle filtering and should be
        possible to implement for every reasonable model.
        """
        raise NotImplementedError("PDF not implemented.")

    @property
    @abstractmethod
    def ndim(self):
        """
        """
        raise NotImplementedError
